@@ -1,6 +1,5 @@
 /*
  * xinstall.c
- * $Id$
  * Copyright (c) 1987, 1993
  *	The Regents of the University of California.  All rights reserved.
  *
@@ -48,6 +47,14 @@
 #include <sys/cdefs.h>
 #endif
 
+/* For clonefile */
+#ifdef HAVE_SYS_ATTR_H
+#include <sys/attr.h>
+#endif
+#ifdef HAVE_SYS_CLONEFILE_H
+#include <sys/clonefile.h>
+#endif
+
 #include <sys/param.h>
 #include <sys/mman.h>
 #include <sys/mount.h>
@@ -132,6 +139,9 @@ static char *funcname;
 static int safecopy, docompare, dostrip, dobackup, dopreserve, nommap;
 static mode_t mode;
 
+#if defined(HAVE_CLONEFILE) && defined(CLONE_NOOWNERCOPY)
+static int	clone(const char *, const char *, int, char *, size_t);
+#endif
 static int	copy(Tcl_Interp *interp, int, const char *, int, const char *, off_t);
 static int	compare(int, const char *, size_t, int, const char *, size_t);
 static int	create_newfile(Tcl_Interp *interp, const char *, int, struct stat *);
@@ -140,11 +150,12 @@ static int	install(Tcl_Interp *interp, const char *, const char *, u_long, u_int
 static int	install_dir(Tcl_Interp *interp, char *);
 static u_long	numeric_id(Tcl_Interp *interp, const char *, const char *, int *rval);
 static void	strip(const char *);
+static void tempfile_template(const char *, char *, size_t);
 static int	trymmap(int);
 static void	usage(Tcl_Interp *interp);
 
 int
-InstallCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
+InstallCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
 	struct stat from_sb, to_sb;
 	mode_t *set;
@@ -285,11 +296,8 @@ InstallCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *
 	}
 
 	/* must have at least two arguments, except when creating directories */
-	if (objc < 2 && !dodir) {
-		usage(interp);
-		return TCL_ERROR;
-	}
-	else if (dodir && !objc) {
+	if ((objc < 2 && !dodir) ||
+		(dodir && !objc)) {
 		usage(interp);
 		return TCL_ERROR;
 	}
@@ -310,10 +318,7 @@ InstallCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *
 			return TCL_ERROR;
 		}
 		else {
-			char msg[255];
-
-			snprintf(msg, sizeof msg, "%s: chdir(%s)\n", funcname, curdir);
-			ui_info(interp, msg);
+                        ui_info(interp, "%s: chdir(%s)", funcname, curdir);
 		}
 	}
 
@@ -439,12 +444,14 @@ install(Tcl_Interp *interp, const char *from_name, const char *to_name, u_long f
 {
 	struct stat from_sb, temp_sb, to_sb;
 	struct timeval tvb[2];
-	int devnull, files_match, from_fd = 0, serrno, target;
-	int tempcopy, temp_fd, to_fd = 0;
+	int devnull, files_match, from_fd = -1, serrno, target;
+	int tempcopy, temp_fd = -1, to_fd = -1;
 	char backup[MAXPATHLEN], *p, pathbuf[MAXPATHLEN], tempfile[MAXPATHLEN];
-
-        /* message contains function name, two paths and a little bit extra formatting */
-        char msg[MAXPATHLEN * 2 + 32];
+#if defined(HAVE_CLONEFILE) && defined(CLONE_NOOWNERCOPY)
+	const int tryclone = 1;
+#else
+	const int tryclone = 0;
+#endif
 
 	files_match = 0;
 
@@ -494,7 +501,7 @@ install(Tcl_Interp *interp, const char *from_name, const char *to_name, u_long f
 	/* Only copy safe if the target exists. */
 	tempcopy = safecopy && target;
 
-	if (!devnull && (from_fd = open(from_name, O_RDONLY, 0)) < 0) {
+	if ((!tryclone || dostrip || docompare) && !devnull && (from_fd = open(from_name, O_RDONLY, 0)) < 0) {
 		char errmsg[255];
 
 		snprintf(errmsg, sizeof errmsg, "%s: Unable to open: %s, %s",
@@ -527,6 +534,26 @@ install(Tcl_Interp *interp, const char *from_name, const char *to_name, u_long f
 	}
 
 	if (!files_match) {
+	    int done_clone = 0;
+#if defined(HAVE_CLONEFILE) && defined(CLONE_NOOWNERCOPY)
+	    if (tryclone && !devnull && !dostrip) {
+	        done_clone = (clone(from_name, to_name, tempcopy, tempfile, sizeof(tempfile)) == 0);
+	    }
+	    if (done_clone &&
+	        (((to_fd = open(tempcopy ? tempfile : to_name, O_RDONLY, 0)) < 0)
+	        || (!tempcopy && fstat(to_fd, &to_sb) == -1))) {
+                (void)unlink(tempcopy ? tempfile : to_name);
+                done_clone = 0;
+		}
+		if (!done_clone && from_fd < 0 && (from_fd = open(from_name, O_RDONLY, 0)) < 0) {
+            char errmsg[255];
+            snprintf(errmsg, sizeof errmsg, "%s: Unable to open: %s, %s",
+                 funcname, from_name, strerror(errno));
+            Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
+            return TCL_ERROR;
+		}
+#endif
+        if (!done_clone) {
 		if (tempcopy) {
 			to_fd = create_tempfile(to_name, tempfile,
 			    sizeof(tempfile));
@@ -547,16 +574,16 @@ install(Tcl_Interp *interp, const char *from_name, const char *to_name, u_long f
 				Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
 				return TCL_ERROR;
 			}
-			snprintf(msg, sizeof msg, "%s: %s -> %s\n", funcname, from_name, to_name);
-			ui_info(interp, msg);
+			ui_info(interp, "%s: %s -> %s", funcname, from_name, to_name);
 		}
+		} /* !done_clone */
 		if (!devnull) {
-			if (copy(interp, from_fd, from_name, to_fd,
+			if (!done_clone && copy(interp, from_fd, from_name, to_fd,
 			     tempcopy ? tempfile : to_name, from_sb.st_size) != TCL_OK)
 				return TCL_ERROR;
 #if HAVE_COPYFILE
 			if (copyfile)
-				copyfile(from_name, tempcopy ? tempfile : to_name, 0, 0x5);
+				copyfile(from_name, tempcopy ? tempfile : to_name, 0, 0x5); /* 0x5 is COPYFILE_ACL | COPYFILE_XATTR. */
 #endif
 		}
 	}
@@ -650,8 +677,7 @@ install(Tcl_Interp *interp, const char *from_name, const char *to_name, u_long f
 				Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
 				return TCL_ERROR;
 			}
-			snprintf(msg, sizeof msg, "%s: %s -> %s\n", funcname, to_name, backup);
-			ui_info(interp, msg);
+                        ui_info(interp, "%s: %s -> %s", funcname, to_name, backup);
 			if (rename(to_name, backup) < 0) {
 				char errmsg[255];
 
@@ -664,8 +690,7 @@ install(Tcl_Interp *interp, const char *from_name, const char *to_name, u_long f
 				return TCL_ERROR;
 			}
 		}
-		snprintf(msg, sizeof msg, "%s: %s -> %s\n", funcname, from_name, to_name);
-		ui_info(interp, msg);
+                ui_info(interp, "%s: %s -> %s", funcname, from_name, to_name);
 		if (rename(tempfile, to_name) < 0) {
 			char errmsg[255];
 
@@ -851,12 +876,11 @@ compare(int from_fd, const char *from_name UNUSED, size_t from_len,
 	return rv;
 }
 
-/*
- * create_tempfile --
- *	create a temporary file based on path and open it
+/* tempfile_template --
+ * prepare a template filename for use with mktemp/mkstemp.
  */
-static int
-create_tempfile(const char *path, char *temp, size_t tsize)
+static void
+tempfile_template(const char *path, char *temp, size_t tsize)
 {
 	char *p;
 
@@ -868,6 +892,16 @@ create_tempfile(const char *path, char *temp, size_t tsize)
 		p = temp;
 	(void)strncpy(p, "INS@XXXX", &temp[tsize - 1] - p);
 	temp[tsize - 1] = '\0';
+}
+
+/*
+ * create_tempfile --
+ *	create a temporary file based on path and open it
+ */
+static int
+create_tempfile(const char *path, char *temp, size_t tsize)
+{
+	tempfile_template(path, temp, tsize);
 	return (mkstemp(temp));
 }
 
@@ -885,7 +919,6 @@ create_newfile(Tcl_Interp *interp, const char *path, int target, struct stat *sb
 	char backup[MAXPATHLEN];
 	int saved_errno = 0;
 	int newfd;
-	char msg[256];
 
 	if (target) {
 		/*
@@ -909,8 +942,7 @@ create_newfile(Tcl_Interp *interp, const char *path, int target, struct stat *sb
 				return -1;
 			}
 			(void)snprintf(backup, MAXPATHLEN, "%s%s", path, suffix);
-			snprintf(msg, sizeof msg, "%s: %s -> %s\n", funcname, path, backup);
-			ui_info(interp, msg);
+                        ui_info(interp, "%s: %s -> %s", funcname, path, backup);
 			if (rename(path, backup) < 0) {
 				char errmsg[255];
 
@@ -1008,6 +1040,24 @@ copy(Tcl_Interp *interp, int from_fd, const char *from_name, int to_fd, const ch
 	return TCL_OK;
 }
 
+/*
+ * clone --
+ *	create a clone of a file
+ */
+#if defined(HAVE_CLONEFILE) && defined(CLONE_NOOWNERCOPY)
+static int
+clone(const char *from_name, const char *to_name,
+      int use_temp, char *temp_name, size_t tsize)
+{
+    if (use_temp) {
+        tempfile_template(to_name, temp_name, tsize);
+        mktemp(temp_name);
+        to_name = temp_name;
+    }
+    return clonefile(from_name, to_name, CLONE_NOFOLLOW|CLONE_NOOWNERCOPY);
+}
+#endif
+
 /*
  * strip --
  *	use strip(1) to strip the target file
@@ -1044,9 +1094,9 @@ strip(const char *to_name)
 
 /*
  * install_dir --
- *	build directory heirarchy
+ *	build directory hierarchy
  */
-int
+static int
 install_dir(Tcl_Interp *interp, char *path)
 {
 	char *p;
@@ -1068,10 +1118,7 @@ install_dir(Tcl_Interp *interp, char *path)
 					return TCL_ERROR;
 				}
 				else {
-					char msg[255];
-
-					snprintf(msg, sizeof msg, "%s: mkdir %s\n", funcname, path);
-					ui_info(interp, msg);
+                                        ui_info(interp, "%s: mkdir %s", funcname, path);
 				}
 			} else if (!S_ISDIR(sb.st_mode)) {
 				char errmsg[255];
@@ -1115,7 +1162,7 @@ usage(Tcl_Interp *interp)
  * trymmap --
  *	return true (1) if mmap should be tried, false (0) if not.
  */
-int
+static int
 #ifdef MFSNAMELEN
 trymmap(int fd)
 #else
