@@ -1,5 +1,4 @@
 /*
- * $Id$
  * Copyright (c) 1987, 1993
  *	The Regents of the University of California.  All rights reserved.
  *
@@ -28,4 +27,4 @@
  * SUCH DAMAGE.
  */
 
-int InstallCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
+int InstallCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]);
